"""
connect action

Connects and authenticates OAuth2.

"""

import logging

from ..action import Action

class Connect(Action):
    """Connects and authenticates OAuth accounts."""

    # name of action
    name = "connect"

    @classmethod
    def add_arguments(cls, parser):
        """Add arguments to the parser for collection in app.args.

        Args:
            parser:
                `argparse.ArgumentParser`. Parser.
                Arguments added here are server on
                self.args.
        """

        parser.add_argument(
            '--project',
            action='store',
            nargs='?',
            const='',
            dest='asana_project',
            help="asana project id.",
            )

        parser.add_argument(
            '--repo',
            action='store',
            nargs='?',
            const='',
            dest='github_repo',
            help="github repository id.",
            )
        pass

    def run(self):
        app = self.app

        logging.info("connected ok.")


"""
Defines an Action.
"""

def get_subclasses(c):
    """Gets the subclasses of a class."""
    subclasses = c.__subclasses__()
    for d in list(subclasses):
        subclasses.extend(get_subclasses(d))
    return subclasses

class Action(object):
    """A ToolApp Action."""

    # name of action
    name = "unnamed"

    def __init__(self, args, app):
        self.args = args
        self.app = app

    def run(self):
        raise NotImplementedError(
            "run() is not implemented in %s" % self.__class__.__name__)

    @classmethod
    def add_arguments(cls, parser):
        """Add arguments to the parser for collection in app.args.

        Args:
            parser:
                `argparse.ArgumentParser`. Parser.
                Arguments added here are server on
                self.args.
        """

        parser.add_argument(
            '-as-api', '--asana-api',
            action='store',
            nargs='?',
            const='',
            dest='asana_api',
            help="[setting] asana api key.",
            )

        parser.add_argument(
            '-gh-api', '--github-api',
            action='store',
            nargs='?',
            const='',
            dest='github_api',
            help="[setting] github api token.",
            )

        parser.add_argument(
            '--first-issue',
            type=int,
            action='store',
            nargs='?',
            const='',
            help="[setting] only sync issues [FIRST_ISSUE] and above"
            )

    def get_repo_and_project(self):
        """Returns repository and project."""
        app = self.app

        # Get repo
        repo = app.data.apply('github-repo', app.args.github_repo,
            app.prompt_repo,
            on_load=app.github.get_repo,
            on_save=lambda r: r.id
            )

        assert repo, "repository not found."

        # Get project
        project = app.data.apply('asana-project', app.args.asana_project,
            app.prompt_project,
            on_load=app.asana.projects.find_by_id,
            on_save=lambda p: p['id']
            )

        assert project, "project not found."

        # Set first issue
        first_issue = app.data.apply('first-issue', app.args.first_issue,
            "set the first issue to sync with [1 for new repos]",
            on_save=int)

        assert first_issue
        assert first_issue >= 0, "issue must be positive"

        app.sync_data()

        return repo, project

    @classmethod
    def iter_actions(cls):
        """Iterates over new instances of Actions."""

        for sub_class in get_subclasses(cls):
            yield sub_class

# Import all actions!
from .actions import *



"""
Github.com / Asana.com issue tool.
Synchronises creation of github issues and asana tasks.
"""

__VERSION__ = "0.2.8"

from .tool import ToolApp

def tool_app():
    """Starts the tool."""
    app = ToolApp(version=__VERSION__)
    exit(app.exit_code)


# -*- coding: utf-8 -*-
"""
transport queue handlers

Provides an interface for multiprocessing transport queues.

"""

import logging
import re
import multiprocessing
import Queue
import urllib3
import certifi

from asana import Client
from asana import error as asana_errors
from github import Github

import tool

mem = multiprocessing.Manager()

data = mem.dict()
"""Transient data about github and asana repo."""

shutdown_event = mem.Event()
"""Shutdown event"""

queue = mem.Queue()
"""Multiprocessing transport queue."""

settings_queue = mem.Queue()
"""Multprocessing queue for updating settings."""

processes = []
"""Contains running workers."""

ASANA_SECTION_RE = re.compile(r'## Asana Tasks:\s+(.*#(\d{12,}))+', re.M)
"""Regular exprsssion to catch malformed data due to too many tasks."""

class TransportWorker(object):

    """Represents a single thread worker that responds to a queue of tasks.
    """

    def __init__(self, settings):
        self.settings = settings
        self.asana = Client.basic_auth(self.settings['api-asana'])
        self.asana_me = self.asana.users.me()
        self.github = Github(self.settings['api-github'])
        self.github_user = self.github.get_user()

    def get_repo(self):
        return self.github.get_repo(data['github-repo'])

    def run(self):

        while True:

            if shutdown_event.is_set():
                logging.debug("got shutdown signal")
                break

            try:
                packet = queue.get(timeout=5)
            except Queue.Empty:
                continue

            # Handle packet
            packet_task = packet.pop('task')
            method = getattr(self, packet_task, None)
            if not method:
                raise Exception("Packet method '%s' is not supported." %
                                packet_task)

            logging.debug("running packet: %s", packet_task)
            method(**packet)

    def create_missing_task(self,
                            asana_workspace_id,
                            name,
                            assignee,
                            projects,
                            completed,
                            issue_number,
                            issue_html_url,
                            issue_state,
                            issue_body,
                            tasks,
                            labels,
                            label_tag_map):

        """Creates a missing task."""

        task = self.asana.tasks.create_in_workspace(
            asana_workspace_id,
            {
                'name': name,
                'notes': issue_body,
                'assignee': assignee,
                'projects': projects,
                'completed': completed,
            })

        # Announce task git issue
        task_id = task['id']

        put("create_story",
            task_id=task_id,
            text="Git Issue #%d: \n"
                  "%s" % (
                    issue_number,
                    issue_html_url,
                    )
            )

        put("apply_tasks_to_issue",
            tasks=[task_id],
            issue_number=issue_number,
            issue_body=issue_body,
            )

        # Save task to drive
        put_setting("save_issue_data_task",
                    issue=issue_number,
                    task_id=task_id,
                    namespace=issue_state)

        tasks.append(task_id)

        # Sync tags/labels
        put("sync_tags",
            tasks=tasks,
            labels=labels,
            label_tag_map=label_tag_map)

    def add_tag(self, task_id, tag_id):

        try:
            self.asana.tasks.add_tag(task_id=task_id, tag=tag_id)
        except asana_errors.InvalidRequestError:
            logging.warn("warning: bad task %d", task_id)

    def sync_tags(self, tasks, labels, label_tag_map):

        for task_id in tasks:
            tag_ids = []
            try:
                added_tags = 0
                for label in labels:
                    tag_id = label_tag_map.get(label)
                    if not tag_id:
                        continue
                    # if tag_id in tag_ids:
                    #     continue
                    tag_ids.append(tag_id)
                    put("add_tag",
                        task_id=task_id,
                        tag_id=tag_id)
                    added_tags += 1

                if added_tags:
                    put_setting("add_tags_to_task",
                                task_id=task_id,
                                tag_ids=tag_ids)

            except asana_errors.InvalidRequestError:
                logging.warn("warning: bad task %d", task_id)

    def create_story(self, task_id, text):

        self.asana.stories.create_on_task(task_id,
                                          { 'text': text })

    def issue_edit(self, issue_number, body):

        repo = self.get_repo()
        issue = repo.get_issue(issue_number)
        issue.edit(body=body)


    def apply_tasks_to_issue(self, tasks, issue_number, issue_body):
        """Applies task numbers to an issue."""
        issue_body = issue_body
        task_numbers = format_task_numbers_with_links(tasks)
        if task_numbers:
            new_body = ASANA_SECTION_RE.sub('', issue_body)
            new_body = new_body + "\n## Asana Tasks:\n\n%s" % task_numbers
            put("issue_edit",
                issue_number=issue_number,
                body=new_body)
            return new_body

        return issue_body

    def update_task(self, task_id, params):
        self.asana.tasks.update(task_id, params)


def run_worker(settings):
    try:
        worker = TransportWorker(settings)
        worker.run()
    except:
        shutdown_event.set()
        raise

def put(task, **kwargs):
    kwargs['task'] = task
    queue.put(kwargs)

def put_setting(task, **kwargs):
    """Pushes a setting to the queue."""
    kwargs['task'] = task
    settings_queue.put(kwargs)

def flush(callback=None):
    """Waits until queue is empty."""

    while True:
        if shutdown_event.is_set():
            return

        if callable(callback):
            callback()

        try:
            item = queue.get(timeout=1)
            queue.put(item)  # put it back, we're just peeking.
        except Queue.Empty:
            return

def issue_edit(issue, **kwargs):
    """Saves an issue"""
    put("issue_edit",
        issue_number=issue.number,
        **kwargs)

def task_create(asana_workspace_id, name, notes, assignee, projects,
                completed, **kwargs):
    """Creates a task"""
    put("task_create",
        asana_workspace_id=asana_workspace_id,
        name=name,
        notes=notes,
        assignee=assignee,
        projects=projects,
        completed=completed,
        **kwargs)

def start(app):

    logging.debug("Starting multiprocess workers")
    for _ in range(multiprocessing.cpu_count()):
        process = multiprocessing.Process(target=run_worker,
                                          kwargs={
                                            'settings': app.settings.data,
                                          })
        process.start()
        processes.append(process)

def shutdown():
    logging.debug("Shutting down transporter")

    shutdown_event.set()
    for p in processes:
        p.join()

def is_shutdown():
    """Returns True if the app is requesting a global shutdown."""
    return shutdown_event.is_set()

def iter_settings():
    """Yields items from the settings queue."""

    while True:
        try:
            item = settings_queue.get(timeout=1)
            yield item
        except Queue.Empty:
            return

def format_task_numbers_with_links(tasks):
    """Returns formatting for the tasks section of asana."""

    project_id = data.get('asana-project', None)

    def _task_format(task_id):
        if project_id:
            asana_url = tool.ToolApp.make_asana_url(project_id, task_id)
            return "[#%d](%s)" % (task_id, asana_url)
        else:
            return "#%d" % task_id

    return "\n".join([_task_format(tid) for tid in tasks])



"""
Maintains JSON based data file manipulated as a dictionary.
"""

import json

class JSONData(object):

    def __init__(self, filename, args, version):
        """
        Args:
            filename:
                Filename for database.
            args:
                Program arguments.
            version:
                Version of file.
        """
        self.args = args
        self.version = version

        self.filename = filename

        try:
            with open(self.filename, 'rb') as file:
                self.data = json.load(file)
        except IOError:
            self.data = {}

    def assert_version(self):
        """Asserts that the version and data file exists."""

        if not self.has_key('version'):
            raise Exception("`asana-hub connect` must be run first.")

        self.data['version'] = self.version

    def save(self):
        """Save data."""

        with open(self.filename, 'wb') as file:
            self.prune()
            self.data['version'] = self.version
            json.dump(self.data,
                file,
                sort_keys=True, indent=2)

    def __setitem__(self, key, value):
        """Set a value by key."""
        self.data[key] = value

    def __getitem__(self, key):
        """Get a value by key."""
        return self.data[key]

    def prune(self, data=None):
        if data is None:
            data = self.data

        empty_keys = [k for k, v in data.iteritems() if not v]
        for k in empty_keys:
            del data[k]

        for v in data.values():
            if isinstance(v, dict):
                self.prune(data=v)

    def apply(self, key, value, prompt=None,
        on_load=lambda a: a, on_save=lambda a: a):
        """Applies a setting value to a key, if the value is not `None`.

        Returns without prompting if either of the following:
            * `value` is not `None`
            * already present in the dictionary

        Args:
            prompt:
                May either be a string to prompt via `raw_input` or a
                method (callable) that returns the value.

            on_load:
                lambda. Value is passed through here after loaded.

            on_save:
                lambda. Value is saved as this value.
        """

        # Reset value if flag exists without value
        if value == '':
            value = None
            if key and self.data.has_key(key): del self.data[key]

        # If value is explicitly set from args.
        if value is not None:
            value = on_load(value)
            if key: self.data[key] = on_save(value)
            return value

        elif not key or not self.has_key(key):
            if callable(prompt):
                value = prompt()
            elif prompt is not None:
                value = raw_input(prompt + ": ")

            if value is None:
                if self.data.has_key(key): del self.data[key]
                return None

            self.data[key] = on_save(value)
            return value

        return on_load(self.data[key])

    def assert_key(self, key):
        assert self.data.get(key), "%s missing from data" % key

    def has_key(self, *args, **kwargs):
        return self.data.has_key(*args, **kwargs)

    def get(self, key, default_value=None):
        try:
            return self.data[key]
        except KeyError:
            if default_value is not None:
                self.data[key] = default_value

            return default_value

